from raimad.ansec import AnSecError
from raimad.ansec import AnSecRadiusError
from raimad.ansec import AnSecRadiusTooManyArgumentsError
from raimad.ansec import AnSecRadiusNotEnoughArgumentsError
from raimad.ansec import AnSecRadiusIncorrectArgumentsError
from raimad.ansec import AnSecThetaError
from raimad.ansec import AnSecThetaTooManyArgumentsError
from raimad.ansec import AnSecThetaNotEnoughArgumentsError
from raimad.ansec import AnSecThetaIncorrectArgumentsError

from raimad.bbox import EmptyBBoxError

#from raimad.cif import CIFExportError
#from raimad.cif import CannotCompileTransformError




import unittest

import numpy as np

import raimad as rai


class TestPolys(unittest.TestCase):

    def test_angle_between(self):
        self.assertAlmostEqual(
            rai.angle_between((0, 0), (10, 0)),
            np.deg2rad(0))

        self.assertAlmostEqual(
            rai.angle_between((0, 0), (100, 0)),
            np.deg2rad(0))

        self.assertAlmostEqual(
            rai.angle_between((10, 0), (100, 0)),
            np.deg2rad(0))

        self.assertAlmostEqual(
            rai.angle_between((10, 0), (-100, 0)),
            np.deg2rad(180))

        self.assertAlmostEqual(
            rai.angle_between((10, 10), (20, 20)),
            np.deg2rad(45))

    def test_polar(self):
        self.assertIsNone(np.testing.assert_array_almost_equal(
            rai.polar(np.deg2rad(60), 2),
            (1, np.sqrt(3))
            ))

        self.assertIsNone(np.testing.assert_array_almost_equal(
            rai.polar(np.deg2rad(-60), 2),
            (1, -np.sqrt(3))
            ))

        self.assertIsNone(np.testing.assert_array_almost_equal(
            rai.polar(np.deg2rad(180), 100),
            (-100, 0)
            ))

    def test_is_rotated(self):
        self.assertTrue(rai.iters.is_rotated(
            'abc',
            'bca',
            ))

        self.assertFalse(rai.iters.is_rotated(
            'abc',
            'aabc',
            ))

        self.assertFalse(rai.iters.is_rotated(
            'abc',
            'cba',
            ))

        self.assertTrue(rai.iters.is_rotated(
            'abcd',
            'bcda',
            ))

        self.assertTrue(rai.iters.is_rotated(
            [1, 2, 3, 4],
            [3, 4, 1, 2],
            ))

        self.assertTrue(rai.iters.is_rotated(
            [
                (1, 2),
                (3, 4),
                (5, 6),
                (7, 8),
                ],
            [
                (5, 6),
                (7, 8),
                (1, 2),
                (3, 4),
                ],
            ))

        self.assertFalse(rai.iters.is_rotated(
            [
                (1, 2),
                (3, 4),
                (5, 6),
                (7, 8),
                ],
            [
                (1, 2),
                (3, 4),
                (7, 8),
                (5, 6),
                ],
            ))


if __name__ == '__main__':
    unittest.main()




from typing import Generator, Callable, Any, TypeVar, ParamSpec, TypeAlias
from enum import Enum
import functools

import numpy as np

import raimad as rai

fullcircle = np.deg2rad(360)
halfcircle = np.deg2rad(180)
quartercircle = np.deg2rad(90)
eigthcircle = np.deg2rad(45)

semicircle = np.deg2rad(180)
demisemicircle = np.deg2rad(90)
hemidemisemicircle = np.deg2rad(45)

class Orientation(Enum):
    POS = 1
    NEG = -1

def angspace(
        start: float,
        end: float,
        num_steps: int = 50,
        orientation: Orientation = Orientation.POS,
        endpoint: bool = True,
        ) -> np.ndarray[float]:
    """
    Angular space:
    construct an array of angles going from `start` to `end`
    in the positive orientation (counterclockwise).
    """
    if orientation is Orientation.NEG:
        start, end = end, start

    while start < 0:
        start += fullcircle

    while end < start:
        end += fullcircle

    return np.linspace(start, end, num_steps, endpoint=endpoint)

def angle_between(p1, p2):
    """
    Angle between two points
    """
    x, y = np.array(p2) - np.array(p1)
    return np.arctan2(y, x)

def polar(arg, mod=1):
    return np.array([
        np.cos(arg),
        np.sin(arg)
        ]) * mod

def is_compo_class(obj):
    return (
        isinstance(obj, type)
        and issubclass(obj, rai.Compo)
        and obj is not rai.Compo
        )

def _custom_base(value: int, glyphs: list[str]):
    """
    Convert an int into a base-n number
    (generator)
    """
    div, mod = divmod(value, len(glyphs))
    if div > 0:
        yield from _custom_base(div, glyphs)
    yield glyphs[mod]

def custom_base(value: int, glyphs: list[str]):
    return ''.join(_custom_base(value, glyphs))


WINGDINGS = [
    f"\033[0;{color}m{symbol} "
    for color in [31, 32, 34, 35, 36]
    for symbol in r"●■◀▶▲▼▚x╚╔╝═╩╗╦╵╷┴┬╰/\╭╮╯∞o8:"
    ]

def wingdingify(value: int):
    """
    Encode an integer with a bunch of symbols.
    """
    return ''.join((
        custom_base(value, WINGDINGS),
        '\033[0m',
        ))

def preload_generator(factory=tuple):
    def preload_generator_inner(generator):
        @functools.wraps(generator)
        def wrapper(*args, **kwargs):
            return factory(generator(*args, **kwargs))
        return wrapper
    return preload_generator_inner


T = TypeVar('T')
P = ParamSpec('P')
Undecorated: TypeAlias = Callable[P, Generator[str, None, None]]
Decorated: TypeAlias = Callable[P, T]
Inner_Decorator: TypeAlias = Callable[[Undecorated], Decorated]
# ARGHHHHH I feel like a C++ programmer

def join_generator(
        string: str,
        post: Callable[[str], T] = lambda x: x,
        ) -> Inner_Decorator:
    def join_generator_inner(
            generator: Undecorated
            ) -> Decorated:

        @functools.wraps(generator)
        def wrapper(
                *args: P.args,
                **kwargs: P.kwargs
                ) -> T:
            return post(string.join(generator(*args, **kwargs)))

        return wrapper

    return join_generator_inner

def midpoint(p1, p2):
    """
    Midpoint between two points
    """
    return (p1 + p2) / 2

# TODO chaining boundpoint actions?

def klay(cifstring):
    """
    Do not use this function.
    This is a helper for me to debug the CIF export process
    until I come up with something better
    """
    import subprocess
    with open('/home/maybetree/tmp/raimad.cif', 'w') as cif:
        cif.write(cifstring)
    #subprocess.Popen(
    #    'flatpak run de.klayout.KLayout /home/maybetree/tmp/raimad.cif'
    #    .split(' '))





"""
graphviz.py

Utilities for generating DOT code
"""

class DOTString(str):
    def _repr_dot_(self):
        return self

def _adjlist2dot(adjlist):
    """
    Convert adjecency list to DOT code
    (generator)
    """

    yield 'digraph D {\n'
    for from_, to in adjlist:
        yield f'\t {from_} -> {to};\n'
    yield '}\n'

def adjlist2dot(adjlist):
    """
    Convert adjecency list to DOT code
    """
    return ''.join(_adjlist2dot(adjlist))

def _make_dot(adjlist, names):
    yield 'digraph D {\n'

    for rout_num, name in names.items():
        yield f'\t {rout_num} [label="{name}"];'

    for from_, to in adjlist:
        yield f'\t {from_} -> {to};\n'

    yield '}\n'

def make_dot(adjlist, names):
    return ''.join(_make_dot(adjlist, names))




"""
Operations on affine matrices
"""

import numpy as np
import numpy.typing as _

import raimad as rai

def rotate(angle: float) -> 'rai.typing.Affine':
    return np.array([
        [np.cos(angle), -np.sin(angle), 0],
        [np.sin(angle), np.cos(angle), 0],
        [0, 0, 1],
        ])

def move(x: float, y: float) -> 'rai.typing.Affine':
    return np.array([
        [1, 0, x],
        [0, 1, y],
        [0, 0, 1],
        ])

def scale(x: float, y: float) -> 'rai.typing.Affine':
    return np.array([
        [x, 0, 0],
        [0, y, 0],
        [0, 0, 1],
        ])

def around(matrix: 'rai.typing.Affine', x: float, y: float) -> 'pc.typing.Affine':
    to_origin = move(-x, -y)
    from_origin = move(x, y)

    return from_origin @ matrix @ to_origin

def get_translation(matrix: 'rai.typing.Affine') -> np.typing.NDArray[np.float64]:
    """
    Given an affine matrix, return the corresponding translation.
    Written by ChatGPT
    """
    return matrix[:2, 2]

def get_scale(matrix: 'rai.typing.Affine') -> tuple[np.float64, np.float64]:
    """
    Given an affine matrix, return the corresponding scale.
    Written by ChatGPT
    """
    scale_x = np.linalg.norm(matrix[:, 0])
    scale_y = np.linalg.norm(matrix[:, 1])
    return scale_x, scale_y

def get_shear(matrix: 'rai.typing.Affine') -> float:
    """
    Given an affine matrix, return the corresponding shear.
    Written by ChatGPT
    """
    scale_x, scale_y = get_scale(matrix)
    shear = float(np.dot(matrix[:, 0], matrix[:, 1]) / (scale_x * scale_y))
    return shear

def get_rotation(matrix: 'rai.typing.Affine') -> float:
    """
    Given an affine matrix, return the corresponding rotation
    Written by ChatGPT
    """
    return float(np.arctan2(matrix[1, 0], matrix[0, 0]))

def transform_xyarray(
        matrix: 'rai.typing.Affine',
        xyarray: 'rai.typing.Poly | pc.typing.PolyArray'
        ) -> 'rai.typing.Poly | pc.typing.PolyArray':
    """
    Apply transformation to xyarray and return new transformed xyarray
    """
    if len(xyarray) == 0:
        return xyarray

    if not isinstance(xyarray, np.ndarray):
        # TODO instead of dynamic hacks like this,
        # grind through with a static checker
        xyarray = np.array(xyarray)

    homogeneous = np.hstack((
        xyarray,
        np.ones((xyarray.shape[0], 1)),
        ))
    transformed = np.dot(homogeneous, matrix.T)
    euclidean: np.typing.NDArray[np.float64] = \
        transformed[:, :2] / transformed[:, 2].reshape(-1, 1)

    return euclidean

def transform_point(
        matrix: 'rai.typing.Affine',
        point: 'rai.typing.Point'
        ) -> 'rai.typing.Point':
    """
    Apply transformation to point and return new transformed point
    """
    homogeneous = np.append(point, 1)
    transformed = np.dot(homogeneous, matrix.T)
    cartesian = transformed[:2] / transformed[2]

    return np.array(cartesian)





"""
Collection of one-size-fits-all wisdoms and questionable quotations
"""

fortunes = [

    # Technology #

    'The plural of Regex is "Regrets". ',

    'Why make it simple and efficient, '
    'when you can make it a huge overcomplicated mess? ',

    '''"Stable"? That's for horses. '''

    'Roses are red, violets are blue. '
    'Unterminated string literal at line 32. ',

    'A computer is much like air conditioning: '
    'useless the moment you open Windows. ',

    'We built our world without walls and fences, '
    "so naturally we don't need Windows and Gates. ",

    "object: v. to feel strong distaste; to oppose firmly.\n"
    "See also: object-oriented programming. ",

    "Just a few hours of trial and error can save "
    "countless minutes of reading the documentation. ",

    "Printing is not supported on this printer. ",

    "Required qualifications: minimal Java experience (ideally none)",

    "Configuring login daemon...\n"
    "Starting login daemon...\n"
    "Exorcising login daemon...\n"
    "System operational.",

    # Education #

    '"The doors are wide open, '
    'but we still see people crashing into the doorframe.\n" '
    '    -- NvdM on IC exams',

    'In order to save energy and help meet our goal of a carbon-neutral '
    'campus by 2025, the light at the end of the tunnel has been turned off. ',

    "Technical university: technically, it's a university. ",

    "Theoretical physicists: in theory, they're good at physics. ",

    "Don't try; just do. ",

    "If you can't, then don't. ",

    "Pay attention, I know exactly what I am talking about. ",

    "Just because [pipelining] wasn't covered in the lectures "
    "doesn't meant it won't be on the exam. ",

    "Learning by doing: I do, you learn. ",

    "There are no stupid questions, "
    "only stupid people. ",

    # Politics #

    "Abuse of authority comes as no surprise. ",

    "Money doesn't bring happiness, but neither does poverty. ",

    '"Somehow the potential to make a lot of money has become more valuable '
    'then making a lot of money lol. " -- u/h3J1e on shareholders',

    "Elect a clown, expect a circus. ",

    "Points of discussion: "
    "construction of a shed, and construction of communism. "
    "Due to lumber shortages, it was decided to skip "
    "straight to the second point. ",

    'Now is the time to save face. We hope there will be time to '
    'save other parts of the body as well. ',

    "We wanted the best, but it turned out like always. ",

    "No matter what we try to develop, we always end up with either "
    "the communist party or the Khalashnikov rifle.",

    "You can't hug a child with nuclear arms. ",

    'Experimental missile nicknamed "Civil Servant": '
    "won't work and they can't fire it",

    "We have thoroughly addressed all concerns, from A to B. ",

    "He is a man of logic and should not be taken seriously. ",

    "That is my opinion. Do not ask me why. ",

    "- A choice of what? \n"
    "- A choice of watt? \n"
    "- Yes, a choice of what? \n"
    "- No, the watts are exactly the same, "
    "the choice is who you buy them from. ",

    "- How long are you in for?\n"
    "- Three years\n"
    "- What for?\n"
    "- For nothing!\n"
    "- Do not try to play us for fools, "
    'everyone knows that "for nothing" is 5 years. ',

    # Malapropisms #

    "Whatever bloats your goat! ",

    # Engineering #

    "Instructions for use: do not. ",

    "Square peg. Round hole. Get it done. ",

    "Emergency phone not installed. "
    "Please do not have an emergency at this location. ",

    'The difference between practice and theory '
    'is often bigger in practice than in theory. ',

    'No keyboard detected. Press F12 to continue.',

    "Forbidden to remove boulder. "
    "Without it, lift does not carry children. ",

    'Anything labeled "this end towards enemy" is dangerous at both ends. ',

    "It works all the time 60% of the time. ",

    # Resilience #

    "Even with everyone's help, you still have to do everything on your own. ",

    "YOU GOTTA BLAME SOMEONE, RIGHT? BLAME YOURSELF",

    "This time it's the same, but also a little bit different. ",

    "It's not a superiority complex if you're actually superior. ",

    "It's not paranoia if someone's actually out to get you. ",

    "The horrors persist, but so do I",

    "New liver, same eagles. ",

    "It is what it is. ",

    "Life is a joke; don't be the punchline. ",

    "Just because you can survive in harsh conditions "
    "doesn't mean you should have to. ",

    "Tip: spice up your panic attack with a harmonica. ",

    # Misc "

    "The best storytellers are those who have a story to tell. ",

    "CAUTION: Gate swings both ways HARD",

    "On one hand, you've got five oranges. On the other hand, you've got "
    "three apples. Conclusion: you have big hands. ",

    "When you are dead, you do not suffer. "
    "The pain is only felt by those around you. "
    "The same happens when you are stupid. ",

    "He's suffered for his art and now it's our turn. ",

    "i may be cringe but youre mean and thats worse",

    "Qaulifications: former child",

    '"Whenever a door closes, another one surely opens"\n'
    '    -- My uncle (wise man, but a terrible cabinet maker)',

    "Present day, present time. ",

    "As above, so below. ",

    "I am not insulting you, I am *describing* you. ",

    "I hope this email finds you before I do. ",

    "To whom it WILL concern: ",

    ]




import os
import raimad as rai

def show(compo: rai.Compo):
    with open('/tmp/raimad_debug.cif', 'w') as f:
        f.write(rai.export_cif(compo))
    #os.system('flatpak run de.klayout.KLayout /tmp/raimad_debug.cif')




import inspect
from typing import Self
from copy import deepcopy

import raimad as rai

class MarksContainer(rai.DictList):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._proxy = None

    def __getattr__(self, name):

        if self._proxy is None:
            return self[name]

        return self._proxy.get_mark(name)

    def _proxy_copy(self, proxy: 'rai.typing.Proxy') -> Self:
        new = type(self)({
            key: val for key, val in self.items()
            })
        new._proxy = proxy
        return new

class SubcompoContainer(rai.DictList):
    def _filter(self, item):
        if isinstance(item, rai.Compo):
            return rai.Proxy(item, _autogen=True)
        elif isinstance(item, rai.Proxy):
            return item
        else:
            raise Exception  # TODO actual exception
        # TODO generally need to standardize runtime checks.

class Compo:
    marks: MarksContainer
    subcompos: SubcompoContainer

    def __init__(self, *args, **kwargs):
        self.geoms = {}
        self.subcompos = SubcompoContainer()
        self.marks = MarksContainer()

        self._make(*args, **kwargs)

    @classmethod
    def partial(cls, **kwargs):
        return rai.Partial(cls, **kwargs)

    def steamroll(self) -> dict:
        """
        Steamroll the entire compo hierarchy into one Geoms dict
        TODO more informative
        """
        geoms = self.geoms.copy()
        for subcompo in self.subcompos.values():
            # TODO override "update" method in geoms container?
            for layer_name, layer_geoms in subcompo.steamroll().items():
                if layer_name not in geoms.keys():
                    geoms[layer_name] = []
                geoms[layer_name].extend(layer_geoms)
        return geoms

    def final(self):
        return self

    def depth(self):
        return 0

    def descend(self):
        yield self

    def descend_p(self):
        return
        yield

    def proxy(self):
        return rai.Proxy(self)

    def copy(*args, **kwargs):
        """
        Don't copy components, copy proxies
        """
        return NotImplemented

    def walk_hier(self):
        yield self
        for subcompo in self.subcompos.values():
            yield from subcompo.walk_hier()

    # Transform functions #
    # TODO for all transforms
    #def scale(self, factor):
    #    return rai.Proxy(
    #        self,
    #        transform=rai.Transform().scale(factor)
    #        )

    #def movex(self, factor):
    #    return rai.Proxy(
    #        self,
    #        transform=rai.Transform().movex(factor)
    #        )

    #def movey(self, factor):
    #    return rai.Proxy(
    #        self,
    #        transform=rai.Transform().movey(factor)
    #        )

    #def rotate(self, angle):
    #    return rai.Proxy(
    #        self,
    #        transform=rai.Transform().rotate(angle)
    #        )

    # lmap function #
    #def __matmul__(self, what):
    #    if isinstance(what, dict | str):  # TODO lmap shorthand type
    #        return rai.Proxy(
    #            self,
    #            lmap=what
    #            )
    #    elif isinstance(what, rai.Transform):
    #        return rai.Proxy(
    #            self,
    #            transform=what
    #            )
    #    else:
    #        raise Exception()  # TODO

    # mark functions #
    def set_mark(self, name, point):
        # TODO this is a boundpoint but not actually a boundpoint?
        self.marks[name] = rai.BoundPoint(point, None)

    def get_mark(self, name):
        return self.marks[name]

    # bbox functions #
    @property
    def bbox(self):
        bbox = rai.BBox()
        for geoms in self.steamroll().values():
            for geom in geoms:
                bbox.add_xyarray(geom)
        return bbox

    def __init_subclass__(cls):
        _class_to_dictlist(cls, 'Marks', rai.Mark)
        _class_to_dictlist(cls, 'Layers', rai.Layer)
        _class_to_dictlist(cls, 'Options', rai.Option)

        for param in inspect.signature(cls._make).parameters.values():
            if param.name not in cls.Options.keys():
                # TODO unannotated
                continue

            cls.Options[param.name].annot = rai.Empty
            cls.Options[param.name].default = rai.Empty

            if param.default is not inspect._empty:
                cls.Options[param.name].default = param.default

            if param.annotation is inspect._empty:
                if param.default is not inspect._empty:
                    cls.Options[param.name].annot = type(param.default)
            else:
                cls.Options[param.name].annot = param.annotation

    # Condemned method, I don't like it
    #def subcompo(self, compo, name: str | None = None):
    #    proxy = rai.Proxy(compo)
    #    if name is None:
    #        self.subcompos.append(proxy)
    #    else:
    #        # TODO runtime check for subcompo reassignment?
    #        self.subcompos[name] = proxy
    #    return proxy

    def auto_subcompos(self, locs=None):
        """
        Automatically add all proxies defined in whatever function you
        call this from as subcompos using some arcane stack inspection
        hackery.
        """
        # Get all local variables in the above frame
        locs = locs or inspect.stack()[1].frame.f_locals

        for name, obj in locs.items():
            if (
                    isinstance(obj, rai.Proxy)
                    and name != 'self'
                    and not name.startswith('_')
                    ):
                # TODO forbid adding compo directly as subcompo,
                # only proxy, instead of doing it automatically.
                self.subcompos[name] = obj

    def __str__(self):
        return (
            "<"
            f"{type(self).__name__} at {rai.wingdingify(id(self))} "
            ">"
            )

    def __repr__(self):
        return self.__str__()

    def _repr_svg_(self):
        """
        Make svg representation of component.
        This is called by jupyter and raimark
        """
        return rai.export_svg(self)

def _class_to_dictlist(cls, attr, wanted_type):
    if not hasattr(cls, attr):
        setattr(cls, attr, rai.DictList())
        return

    new_list = rai.DictList()
    for name, annot in getattr(cls, attr).__dict__.items():
        if not isinstance(annot, wanted_type):
            continue
        annot.name = name
        new_list[name] = annot

    setattr(cls, attr, new_list)




import ast

class Viol():
    """
    Viol: Checker violation.
    This is a base class that can be used to catch
    all checker violtations.
    """

    def __init__(self, line, col=0, **context):
        if isinstance(line, ast.AST):
            self.line = line.lineno
            self.col = line.col_offset
        else:
            self.line = line
            self.col = col
        self.context = context

    def flake8(self):
        desc = self.desc.format(**self.context)
        name = type(self).__name__
        return f"{name} {desc}"

    def __str__(self):
        desc = self.desc.format(**self.context)
        name = type(self).__name__
        return f"<{name} at line {self.line}: {desc}>"

    def __repr__(self):
        return self.__str__()

class MarksViol(Viol):
    """
    MarksViol: Marks-related Checker violation.
    This is a base class that can be used to catch
    all marks-related checker violtations.
    All marks-related checker violations have
    a code in the form RAI4XX.
    """

class LenientViol(Viol):
    """
    LenientViol: Checker violation in Lenient mode.
    This is a base class that can be used to catch
    all checker violations that are raised in lenient mode
    only.
    Violations that are raised in stricter modes
    do not inherit from this class.
    """

class AuthoritarianViol(Viol):
    """
    """

class RAI412(MarksViol, LenientViol):
    desc = "Mark `{mark}` assigned multiple times at lines {lines}"

class RAI442(MarksViol, AuthoritarianViol):
    desc = "Assignment to undeclared mark `{mark}`"




"""
empty.py
This file contains the EmptyType class and the Empty object
"""

from typing import Self, ClassVar

class EmptyType:
    """
    Empty: an object representing an empty field.
    This can be used similar to `None`,
    except in cases where `None` has a different meaning.
    For example, compo options by default have their
    `browser_default` property set to Empty,
    indicating that there is no default value to use
    in the component browser.
    Setting it to None would mean that the default value is None.

    This is a singleton class.
    """
    instance: ClassVar[Self]

    def __new__(cls) -> 'EmptyType':
        if not hasattr(cls, 'instance'):
            cls.instance = super(EmptyType, cls).__new__(cls)
        return cls.instance

    def __bool__(self) -> bool:
        return False

    def __str__(self) -> str:
        return "<Empty>"

    def __repr__(self) -> str:
        return "<Empty>"


Empty: EmptyType = EmptyType()




from typing import Any, Self
import numpy as np

import raimad as rai

# see
# https://numpy.org/doc/stable/user/basics.subclassing.html
class BoundPoint(np.ndarray[Any, np.dtype[np.float64]]):

    _proxy: 'rai.typing.Proxy | None'

    def __new__(
            cls,
            input_array: 'rai.typing.Point',
            proxy: 'rai.typing.Proxy'
            ) -> Self:
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj._proxy = proxy
        # Finally, we must return the newly created object:
        return obj

    @property
    def proxy(self) -> 'rai.typing.Proxy':
        if self._proxy is None:
            # TODO exception type
            raise Exception('Unbound BoundPoint')

        return self._proxy

    def __array_finalize__(self, obj: np.typing.NDArray[np.float64] | None) -> None:
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        self._proxy = getattr(obj, 'proxy', None)

    def to(self, point: 'rai.typing.Point') -> 'pc.typing.Proxy':
        self.proxy.transform.move(
            point[0] - self[0],
            point[1] - self[1],
            )
        return self.proxy

    def rotate(self, angle: float) -> 'rai.typing.Proxy':
        self.proxy.transform.rotate(
            angle,
            self[0],
            self[1]
            )
        return self.proxy

    def move(self, x: float, y: float) -> 'rai.typing.Proxy':
        """
        `move`ing a boundpoint is functionally identical to `move`ing
        the proxy itself, so there is no reason to call this explicitly.
        However, this function is still defined, so that you can use
        `move` in a chain of actions on a boundpoint
        """
        self.proxy.transform.move(x, y)
        return self.proxy

    # TODO the rest of the functions




import ast
import inspect

import raimad as rai

def write_parents(tree):
    """
    Walk a tree and add a `parent` attribute to every node
    (that points to the direct parent)
    and an `ancestors` attribute that is a list of ancestors
    from closest to latest.
    """
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node

    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):

            child.ancestors = []
            ancestor = child
            while 1:
                try:
                    ancestor = ancestor.parent
                except AttributeError:
                    break

                child.ancestors.append(ancestor)

#def __matches(match, case):
#    if match == case:
#        return True
#
#    if not hasattr(case, '__match_args__'):
#        return False
#
#    if not isinstance(match, type(case)):
#        return False
#
#    for arg in case.__match_args__:
#        try:
#            attr_match = getattr(match, arg)
#            attr_case = getattr(case, arg)
#        except AttributeError:
#            continue
#
#        if not __matches(attr_match, attr_case):
#            return False
#
#    return True

# The name won't actually get scrambled since it's not in a class,
# but this gets the point across that you really shouldn't use this
# outside of this module.
def __matches(match, case):
    exec(
        "match match:\n"
        f"   case {case}:\n"
        "       did_match = True",
        globals(),
        locs := {'match': match}
        )
    return 'did_match' in locs.keys()

def __find(tree, node, multiple=False):
    found = []
    for child in ast.iter_child_nodes(tree):
        if __matches(child, node):
            if multiple:
                found.append(child)
            else:
                return child
    if multiple and found:
        return found

    raise Exception  # TODO

def __find_recursive(tree, node, multiple=False):
    found = []
    for child in ast.walk(tree):
        if __matches(child, node):
            if multiple:
                found.append(child)
            else:
                return child
    if multiple and found:
        return found
    raise Exception  # TODO

def _check_compo(compo, tree):
    """
    """
    make_func = __find(tree, """ast.FunctionDef(
        name='_make'
        )""")

    mark_assigns = __find_recursive(tree, """ast.Attribute(
        value=ast.Attribute(
            value=ast.Name(id='self'),
            attr='marks',
            ),
        )""", True)

    redundancy = {}
    for assign in mark_assigns:
        mark_name = assign.attr

        if mark_name not in [mark.name for mark in compo.Marks.values()]:
            yield rai.RAI442(assign, mark=mark_name)

        if mark_name not in redundancy.keys():
            redundancy[mark_name] = []
        redundancy[mark_name].append(assign)

    # TODO how to deal with multiple asignment
    # that's not actually multiple asignment
    # because it's in a mutually exclusive `if`?
    #for mark_name, nodes in redundancy.items():
    #    if len(nodes) > 1:
    #        for node in nodes:
    #            yield rai.RAI412(node, mark=mark_name, lines=[
    #                node.lineno for node in nodes
    #                ])


def check_compo(compo):
    code = inspect.getsource(compo)
    root = ast.parse(code)
    tree = __find(root, """ast.ClassDef()""")
    write_parents(tree)
    yield from _check_compo(compo, tree)

def check_module(tree):
    compiled = compile(tree, '<string>', 'exec')
    exec(
        compiled,
        globals(),  # TODO is this correct?
        locs := {}
        )

    for compo_node in __find(tree, "ast.ClassDef()", True):
        compo = locs.get(compo_node.name, None)
        if compo is None:
            continue
        if not isinstance(compo, type):
            continue
        if not issubclass(compo, rai.Compo):
            continue
        yield from _check_compo(compo, compo_node)

class Flake8Checker:
    name = __name__
    version = 0

    def __init__(self, tree: ast.AST) -> None:
        self._tree = tree

    def run(self):
        if not hasattr(self._tree, 'body'):
            return
        for toplevel in self._tree.body:
            match toplevel:
                case ast.Expr(value=ast.Constant(value='pc_checkme')):
                    break
        else:
            return

        for viol in check_module(self._tree):
            yield viol.line, viol.col, viol.flake8(), type





from typing import Any
from copy import copy
import raimad as rai

class Partial:
    compo_cls: 'rai.typing.CompoClass'

    def __init__(self, compo_cls: 'rai.typing.CompoClass', **kwargs: Any) -> None:
        self.kwargs = kwargs
        self.compo_cls = compo_cls

    def __call__(self, **kwargs: Any) -> 'rai.typing.RealCompo':
        kwargs2 = copy(self.kwargs)
        kwargs2.update(kwargs)
        return self.compo_cls(**kwargs2)




import unittest

import raimad as rai

class Unannotated(rai.Compo):
    class Marks:
        square_corner = rai.Mark('corner of the square')

    def _make(self):
        self.marks.triangle_corner = (20, 40)
        self.marks.square_center = (5, 5)
        self.marks.square_corner = (10, 10)

class Annotated(rai.Compo):
    class Marks:
        square_corner = rai.Mark('corner of the square')
        square_center = rai.Mark('corner of the square')
        triangle_corner = rai.Mark('corner of the square')

    def _make(self):
        self.marks.triangle_corner = (20, 40)
        self.marks.square_center = (5, 5)
        self.marks.square_corner = (10, 10)

class Reassign(rai.Compo):
    class Marks:
        square_corner = rai.Mark('corner of the square')

    def _make(self):
        self.marks.square_corner = (5, 5)
        self.marks.square_center = (1, 1)
        print('useless statement')
        self.marks.square_corner = (6, 6)

def same_viols(actual, expected):
    while len(actual) and len(expected):
        for i, (line, viol_class) in enumerate(expected):
            matched = False
            for j, viol in enumerate(actual):
                if isinstance(viol, viol_class) and viol.line == line:
                    del expected[i]
                    del actual[j]
                    matched = True
                    break
            if matched:
                break
        if not matched:
            return False
    if len(actual) or len(expected):
        return False
    return True


class TestCheckMarkNames(unittest.TestCase):

    def test_lint_annotated(self):
        viols = list(rai.check_compo(Annotated))
        self.assertEqual(len(viols), 0)

    def test_lint_unannotated(self):
        viols = list(rai.check_compo(Unannotated))
        self.assertTrue(same_viols(viols, [
            (6, rai.RAI442),
            (7, rai.RAI442),
            ]))

#    def test_lint_reassign(self):
#        viols = list(rai.check_compo(Reassign))
#        self.assertTrue(same_viols(viols, [
#            (6, rai.RAI412),
#            (7, rai.RAI442),
#            (9, rai.RAI412),
#            ]))


if __name__ == '__main__':
    unittest.main()




"""
typing.py
A collection of type hints
"""

from typing import TypeAlias, Sequence, Any
import numpy as np
import raimad as rai

CompoClass: TypeAlias = type[rai.Compo] | rai.Partial
Compo: TypeAlias = rai.Compo | rai.Proxy
RealCompo: TypeAlias = rai.Compo
Proxy: TypeAlias = rai.Proxy
Point: TypeAlias = \
        tuple[float, float] | \
        np.typing.NDArray[np.float64] | \
        rai.BoundPoint
Poly: TypeAlias = Sequence[Point]
PolyArray: TypeAlias = np.typing.NDArray[np.float64]
Polys: TypeAlias = Sequence[Poly]
Geoms: TypeAlias = dict[str, Polys]
Transform: TypeAlias = rai.Transform
BBox: TypeAlias = rai.BBox
Affine: TypeAlias = np.typing.NDArray[np.float64]

LMapShorthand: TypeAlias = None | str | dict[str, str]
#XYarray: TypeAlias = np.ndarray[Any, Point]
#Geoms: TypeAlias = dict[str, list[XYarray]]
LMap: TypeAlias = rai.LMap

# Well, isn't this lovely?
Bool: TypeAlias = bool | np.bool_





from typing import Self, ClassVar
import raimad as rai

class Option:
    def __init__(self, desc: str, browser_default=rai.Empty):
        self.desc = desc
        self.category = type(self)
        self.name = 'this should be set in compo.__init_subclass_'
        self.default = 'this should be set in compo.__init_subclass_'
        self.browser_default = browser_default
        self.annot = 'this should be set in compo.__init_subclass_'

    # These are shorthands which will be assigned at
    # the end of this file.
    # The annotations are here to make mypy shut up
    Functional: ClassVar[type[Self]]
    Geometric: ClassVar[type[Self]]
    Debug: ClassVar[type[Self]]
    Environmental: ClassVar[type[Self]]

class Geometric(Option):
    pass

class Functional(Option):
    pass

class Debug(Option):
    pass

class Environmental(Option):
    pass


Option.Functional = Functional
Option.Geometric = Geometric
Option.Debug = Debug
Option.Environmental = Environmental




import numpy as np
import raimad as rai

class Circle(rai.Compo):
    """
    Circle

    A polygon that approximates a circle,
    defined by the radius and number of points.
    """
    browser_tags = ["builtin", "polygon"]

    class Options:
        radius = rai.Option('Circle radius', browser_default=15)
        num_points = rai.Option('Number of points')

    def _make(self, radius: float, num_points: int = 200):

        self.geoms.update({
            'root': [
                np.array([
                    rai.polar(angle, radius)
                    for angle
                    in np.linspace(0, 2 * np.pi, num_points)
                    ])
                ]
            })

        self.marks.center = (0, 0)




import raimad as rai

class RectWH(rai.Compo):
    """
    RectWH

    A rectangle defined by width and height.
    """
    browser_tags = ["builtin", "polygon"]

    class Options:
        width = rai.Option('Width of rectangle', browser_default=15)
        height = rai.Option('Height of rectangle', browser_default=10)

    def _make(self, width: int, height: int):
        self.width = width
        self.height = height

        self.geoms.update({
            'root': [
                [
                    [- width / 2, - height / 2],
                    [+ width / 2, - height / 2],
                    [+ width / 2, + height / 2],
                    [- width / 2, + height / 2],
                    ]
                ]
            })




from typing import Self, Iterator, Any
from copy import copy

import raimad as rai

class LMap:
    def __init__(self, shorthand: 'rai.typing.LMapShorthand') -> None:
        self.shorthand = shorthand

    def __getitem__(self, name: str) -> str:
        if self.shorthand is None:
            return name

        if isinstance(self.shorthand, str):
            return self.shorthand

        if isinstance(self.shorthand, dict):
            return self.shorthand[name]

        assert False
    # TODO hacky

    def copy(self) -> Self:
        return type(self)(self.shorthand)

    def compose(self, other: Self) -> Self:
        if other.shorthand is None:
            # None lmap, pass everything through
            pass

        elif isinstance(other.shorthand, str):
            # single-string lmap, flatten everything to one layer
            self.shorthand = other.shorthand

        elif isinstance(other.shorthand, dict):
            # TODO better way to do this?
            if isinstance(self.shorthand, dict):
                for self_k, self_val in tuple(self.shorthand.items()):
                    try:
                        self.shorthand[self_k] = other[self_val]
                    except KeyError:
                        pass

            elif self.shorthand is None:
                self.shorthand = copy(other.shorthand)

            elif isinstance(self.shorthand, str):
                self.shorthand = other[self.shorthand]

        else:
            assert False

        return self

class Proxy:
    compo: 'rai.typing.Compo'

    def __init__(self,
                 compo: 'rai.typing.Compo',
                 lmap: 'rai.typing.LMapShorthand' = None,
                 transform: 'rai.typing.Transform | None' = None,
                 _cif_link: bool = False,
                 _autogen: bool = False,
                 ):
        self._cif_linked = False
        self._cif_link = _cif_link
        self._autogen = _autogen
        self.compo = compo
        self.lmap = LMap(lmap)
        self.transform = transform or rai.Transform()

    def steamroll(self) -> 'rai.typing.Geoms':
        return {
            self.lmap[layer]: [
                self.transform.transform_xyarray(geom)
                for geom in geoms
                ]
            for layer, geoms
            in self.compo.steamroll().items()
            }

    def get_flat_transform(self, maxdepth: int = -1) -> 'rai.typing.Transform':
        if maxdepth == 0 or isinstance(self.compo, rai.Compo):
            return self.transform.copy()

        #return (
        #    self.compo.get_flat_transform(maxdepth - 1)
        #    .compose(self.transform)
        #    )

        # TODO huh? copy?
        return self.transform.copy().compose(
            self.compo.get_flat_transform(maxdepth - 1)
            )

    def get_flat_lmap(self, maxdepth: int = -1) -> LMap:
        if maxdepth == 0 or isinstance(self.compo, rai.Compo):
            return self.lmap.copy()

        return self.lmap.copy().compose(
            self.compo.get_flat_lmap(maxdepth - 1)
            )

    @property
    def geoms(self) -> 'rai.typing.Geoms':
        return {
            self.lmap[layer]: [
                self.transform.transform_xyarray(geom)
                for geom in geoms
                ]
            for layer, geoms
            in self.compo.geoms.items()
            }

    @property
    def subcompos(self) -> rai.SubcompoContainer:
        # TODO FrozenSubcompoContainer?
        return rai.SubcompoContainer({
            name: self.copy_reassign(subcompo, _autogen=True)
            for name, subcompo in self.compo.subcompos.items()
            })
        return self.compo.subcompos

    def final(self) -> 'rai.typing.RealCompo':
        return self.compo.final()

    def depth(self) -> int:
        return self.compo.depth() + 1

    def descend(self) -> 'Iterator[rai.typing.Compo]':
        yield self
        yield from self.compo.descend()

    def descend_p(self) -> 'Iterator[rai.typing.Proxy]':
        yield self
        yield from self.compo.descend_p()

    def proxy(self) -> 'rai.typing.Proxy':
        return rai.Proxy(self)

    def walk_hier(self) -> 'Iterator[rai.typing.Proxy]':
        for subcompo in self.compo.walk_hier():
            yield self.copy_reassign(subcompo)

    def copy(self) -> 'rai.typing.Proxy':
        if self.depth() > 1:
            # TODO think about why someone would want to do this
            raise Exception("Copying proxy of depth more than 1")

        return type(self)(
            self.compo,
            self.lmap.shorthand,
            self.transform.copy(),
            )

    #def cifcopy(self):
    #    if self.depth() > 1:
    #        # TODO think about why someone would want to do this
    #        raise Exception("Copying proxy of depth more than 1")

    #    self._cif_linked = True

    #    return type(self)(
    #        self.compo,
    #        copy(self.lmap),
    #        self.transform.copy(),
    #        _cif_link=True
    #        )

    def copy_reassign(
            self,
            new_subcompo: 'rai.typing.Compo',
            _autogen: bool = False,
            ) -> 'rai.typing.Proxy':

        return type(self)(
            new_subcompo,
            self.lmap.shorthand,
            self.transform.copy(),
            _autogen=_autogen,
            )

    def deepcopy(self) -> Any:
        """
        No clue what this is supposed to do
        """
        return NotImplemented
        #return type(self)(
        #    (
        #        self.compo
        #        if isinstance(self.compo, rai.Compo)
        #        else self.compo.copy()
        #        ),
        #    copy(self.lmap),
        #    self.transform.copy(),
        #    )

    def __str__(self) -> str:
        stack = ''.join([
            'ma'[proxy._autogen]
            for proxy in self.descend_p()
            ])
        return (
            "<"
            f"Proxy of {self.final()} at {rai.wingdingify(id(self))} "
            f"stack `{stack}x`"
            ">"
            )

    def __repr__(self) -> str:
        return self.__str__()

    # Transform functions #
    # TODO for all transforms
    # TODO same implementation as in Compo
    # TODO stack another transform or modify self?
    def scale(self, factor: float) -> Self:
        self.transform.scale(factor)
        return self

    def movex(self, factor: float) -> Self:
        self.transform.movex(factor)
        return self

    def movey(self, factor: float) -> Self:
        self.transform.movey(factor)
        return self

    def move(self, x: float, y: float) -> Self:
        self.transform.move(x, y)
        return self

    def hflip(self, x: float = 0) -> Self:
        self.transform.hflip(x)
        return self

    def vflip(self, y: float = 0) -> Self:
        self.transform.vflip(y)
        return self

    def flip(self, x: float = 0, y: float = 0) -> Self:
        self.transform.flip(x, y)
        return self

    def rotate(self, angle: float, x: float = 0, y: float = 0) -> Self:
        self.transform.rotate(angle, x, y)
        return self

    ## lmap function #
    #def __matmul__(self, lmap):
    #    # TODO warn on override lmap?
    #    self.lmap = LMap(lmap)
    #    return self
    #    #return rai.Proxy(
    #    #    self,
    #    #    lmap=lmap
    #    #    )
    def map(self, lmap_shorthand: 'rai.typing.LMapShorthand') -> Self:
        self.lmap = LMap(lmap_shorthand)
        return self

    # mark functions #
    def get_mark(self, name: str) -> rai.BoundPoint:
        return rai.BoundPoint(
            self.transform.transform_point(
                self.compo.get_mark(name)
                ),
            self
            )

    @property
    def marks(self) -> rai.MarksContainer:
        return self.compo.marks._proxy_copy(self)

    # bbox functions #
    # TODO same as compo -- some sort of reuse?
    @property
    def bbox(self) -> 'rai.typing.BBox':
        bbox = rai.BBox(proxy=self)
        for geoms in self.steamroll().values():
            for geom in geoms:
                bbox.add_xyarray(geom)
        return bbox

    # snapping functions #
    def snap_left(self, other: Self) -> Self:
        self.bbox.mid_right.to(other.bbox.mid_left)
        return self

    def snap_right(self, other: Self) -> Self:
        self.bbox.mid_left.to(other.bbox.mid_right)
        return self

    def snap_above(self, other: Self) -> Self:
        self.bbox.bot_mid.to(other.bbox.top_mid)
        return self

    def snap_below(self, other: Self) -> Self:
        self.bbox.top_mid.to(other.bbox.bot_mid)
        return self

    def _repr_svg_(self) -> str:
        """
        Make svg representation of component.
        This is called by jupyter and raimark
        """
        return rai.export_svg(self)




from typing import (
    ItemsView,
     KeysView,
     ValuesView,
     Iterable,
     TypeVar,
     Generic,
     Any,
     )

# This class could also be implemented by deriving from `dict`
# instead of encapsulating it
# (actually, that's how it worked originally),
# but that has no real benefit, at the cost of not being 100%
# sure which methods are implemented and where.

# TODO tests for this class!

T = TypeVar('T')
class DictList(Generic[T]):
    """
    A dict that is also a list and is accesible with attribute syntax!
    """
    _dict: dict[str | int, T]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._dict = dict(*args, **kwargs)

    def __setattr__(self, name: str, value: T) -> None:

        if name.startswith('_'):
            super().__setattr__(name, value)
            return

        if hasattr(self.__class__, name):
            raise Exception  # TODO actual exception
        else:
            self._dict[name] = self._filter(value)

    def __getattr__(self, name: str) -> T:
        return self._dict[name]

    def __getitem__(self, key: str | int) -> T:
        if isinstance(key, int):
            return list(self._dict.values())[key]
        return self._dict.__getitem__(key)

    def __setitem__(self, key: str | int, value: T) -> None:
        if isinstance(key, int):
            list(self._dict.values())[key] = value
        self._dict.__setitem__(key, value)

    def append(self, item: T) -> None:
        self._dict[len(self._dict)] = self._filter(item)

    def extend(self, items: Iterable[T]) -> None:
        for item in items:
            self.append(item)

    # TODO update method

    def __iter__(self) -> None:
        raise NotImplementedError(
            "Iterating over dictlist is ambiguous. "
            "Please use `.keys()`, `.values()`, or `.items()`."
            )

    def items(self) -> ItemsView[str | int, T]:
        return self._dict.items()

    def values(self) -> ValuesView[T]:
        return self._dict.values()

    def keys(self) -> KeysView[str | int]:
        return self._dict.keys()

    def __len__(self) -> int:
        return self._dict.__len__()

    def _filter(self, item: T) -> T:
        """
        This method is run on every item that gets added to the dictlist.
        This method is intended to be overridden by deriving classes.
        For example, SubcompoContainer uses this to wrap every
        raw compo that gets added in a proxy.
        """
        return item




import unittest

import numpy as np

import raimad as rai

class BareGeometric(rai.Compo):
    def _make(self):
        self.geoms.update({
            'root': [
                np.array([
                    [0, 0],
                    [0, 10],
                    [10, 10],
                    [10, 0],
                    ]),
                np.array([
                    [20, 20],
                    [40, 20],
                    [30, 40],
                    ]),
                ]
            })

class BareStructural(rai.Compo):
    def _make(self):
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'upper'},
                )
            )
        self.subcompos.append(
            rai.Proxy(
                BareGeometric(),
                {'root': 'lower'},
                )
            )

class TestCompo(unittest.TestCase):

    def test_bare_geometric(self):
        compo = BareGeometric()
        geom = compo.steamroll()

        # Test one root layer
        self.assertEqual(geom.keys(), {'root'})

        # Test two polys on the layer
        self.assertEqual(len(geom['root']), 2)

    def test_bare_geometric_copied(self):
        compo = BareGeometric()
        self.assertIsNot(compo.geoms, compo.steamroll())

    def test_bare_structural(self):
        compo = BareStructural()
        geom = compo.steamroll()

        # Test one root layer
        self.assertEqual(geom.keys(), {'upper', 'lower'})

        # Test two polys on the top lyaer
        self.assertEqual(len(geom['upper']), 2)

        # Test two polys on the top lyaer
        self.assertEqual(len(geom['lower']), 2)


if __name__ == '__main__':
    unittest.main()




import raimad as rai

class RectWire(rai.Compo):
    """
    RectWire

    A rectangle defined by two points and a thickness.
    """
    browser_tags = ["builtin", "polygon"]

    class Options:
        p1 = rai.Option('First point', browser_default=(0, 0))
        p2 = rai.Option('Second point', browser_default=(10, 10))
        thickness = rai.Option('Thickness', browser_default=2)

    def _make(self, p1, p2, thickness):
        bearing = rai.angle_between(p1, p2)
        step = rai.polar(
            arg=bearing + rai.quartercircle,
            mod=thickness / 2)

        self.geoms.update({
            'root': [
                [
                    p1 + step,
                    p1 - step,
                    p2 - step,
                    p2 + step,
                    ]
                ]
            })




import numpy as np
import raimad as rai

class AnSecError(Exception):
    pass

class AnSecRadiusError(AnSecError):
    pass

class AnSecRadiusTooManyArgumentsError(AnSecRadiusError):
    pass

class AnSecRadiusNotEnoughArgumentsError(AnSecRadiusError):
    pass

class AnSecRadiusIncorrectArgumentsError(AnSecRadiusError):
    pass

class AnSecThetaError(AnSecError):
    pass

class AnSecThetaTooManyArgumentsError(AnSecThetaError):
    pass

class AnSecThetaNotEnoughArgumentsError(AnSecThetaError):
    pass

class AnSecThetaIncorrectArgumentsError(AnSecThetaError):
    pass

class AnSec(rai.Compo):
    """
    Annular Sector

    A polygon approximating an annular sector,
    or, in working men's terms, a "pizza crust".
    There are many possible ways to define an AnSec;
    # TODO examples
    """
    browser_tags = ["builtin", "polygon"]

    class Options:
        r1 = rai.Option("Inner radius", browser_default=18)
        r2 = rai.Option("Outter radius", browser_default=20)
        theta1 = rai.Option("Angle 1", browser_default=np.deg2rad(45))
        theta2 = rai.Option("Angle 2", browser_default=np.deg2rad(90))

    def _make(
            self,
            r1=None,
            r2=None,
            rmid=None,
            dr=None,
            theta1=None,
            theta2=None,
            thetamid=None,
            dtheta=None,
            orientation=None,
            ):

        r1, r2 = self.interpret_radius(r1, r2, rmid, dr)
        theta1, theta2 = self.interpret_theta(
            theta1,
            theta2,
            thetamid,
            dtheta,
            )

        if r1 == r2:  # TODO epsilon
            self.geoms.update({'root': [[]]})
            return

        angspace = rai.angspace(theta1, theta2, orientation=orientation)

        self.geoms.update({
            'root': [
                np.array([
                    rai.polar(arg=angle, mod=radius)
                    for radius, angles in [
                        [r1, angspace],
                        [r2, reversed(angspace)],
                        ]
                    for angle in angles
                    ])
                ]
            })

    @staticmethod
    def interpret_radius(r1, r2, rmid, dr):
        passed = (
            (r1 is not None) << 3 |
            (r2 is not None) << 2 |
            (rmid is not None) << 1 |
            (dr is not None) << 0
            )

        if passed == 0b0000:
            raise AnSecRadiusNotEnoughArgumentsError(
                "No radius specified"
                )

        if passed == 0b0001:
            raise AnSecRadiusNotEnoughArgumentsError(
                "Radius delta specified, but no r1 or r2"
                )

        if passed == 0b0010:
            raise AnSecRadiusNotEnoughArgumentsError(
                "Midradius specified, but no radius delta"
                )

        if passed == 0b0011:
            return rmid - dr / 2, rmid + dr / 2

        if passed == 0b0100:
            raise AnSecRadiusNotEnoughArgumentsError(
                "r2 specified, but no r1 or radius delta"
                )

        if passed == 0b0101:
            return r2 - dr, r2

        if passed == 0b0110:
            raise AnSecRadiusIncorrectArgumentsError(
                "r2 and midradius specified. "
                "What do you want me to do??"
                )

        if passed == 0b0111:
            raise AnSecRadiusTooManyArgumentsError(
                "r2, midradius, and radius delta specified. "
                "You need only two of these!"
                )

        if passed == 0b1000:
            raise AnSecRadiusNotEnoughArgumentsError(
                "Only r1 specified. "
                "Specify radius delta or r2."
                )

        if passed == 0b1001:
            return r1, r1 + dr

        if passed == 0b1010:
            raise AnSecRadiusIncorrectArgumentsError(
                "R1 and midradius specified. "
                "What do you want me to do??"
                )

        if passed == 0b1011:
            raise AnSecRadiusTooManyArgumentsError(
                "r1, midradius, and radius delta specified. "
                "You need only two of these!"
                )

        if passed == 0b1100:
            return r1, r2

        if passed == 0b1101:
            raise AnSecRadiusTooManyArgumentsError(
                "r1, r2, and radius delta specified. "
                "What do you want me to do??"
                )

        if passed == 0b1110:
            raise AnSecRadiusTooManyArgumentsError(
                "r1, r2, and midradius specified. "
                "What do you want me to do??"
                )

        if passed == 0b1111:
            raise AnSecRadiusTooManyArgumentsError(
                "What do you want me to do??"
                )

    @staticmethod
    def interpret_theta(t1, t2, tmid, dt):
        passed = (
            (t1 is not None) << 3 |
            (t2 is not None) << 2 |
            (tmid is not None) << 1 |
            (dt is not None) << 0
            )

        if passed == 0b0000:
            raise AnSecThetaNotEnoughArgumentsError(
                "No Theta specified"
                )

        if passed == 0b0001:
            raise AnSecThetaNotEnoughArgumentsError(
                "Theta delta specified, but no t1 or t2"
                )

        if passed == 0b0010:
            raise AnSecThetaNotEnoughArgumentsError(
                "MidTheta specified, but no Theta delta"
                )

        if passed == 0b0011:
            return tmid - dt / 2, tmid + dt / 2

        if passed == 0b0100:
            raise AnSecThetaNotEnoughArgumentsError(
                "t2 specified, but no t1 or Theta delta"
                )

        if passed == 0b0101:
            return t2 - dt, t2

        if passed == 0b0110:
            raise AnSecThetaIncorrectArgumentsError(
                "t2 and midTheta specified. "
                "What do you want me to do??"
                )

        if passed == 0b0111:
            raise AnSecThetaTooManyArgumentsError(
                "t2, midTheta, and Theta delta specified. "
                "You need only two of these!"
                )

        if passed == 0b1000:
            raise AnSecThetaNotEnoughArgumentsError(
                "Only t1 specified. "
                "Specify Theta delta or t2."
                )

        if passed == 0b1001:
            return t1, t1 + dt

        if passed == 0b1010:
            raise AnSecThetaIncorrectArgumentsError(
                "t1 and midTheta specified. "
                "What do you want me to do??"
                )

        if passed == 0b1011:
            raise AnSecThetaTooManyArgumentsError(
                "t1, midTheta, and Theta delta specified. "
                "You need only two of these!"
                )

        if passed == 0b1100:
            return t1, t2

        if passed == 0b1101:
            raise AnSecThetaTooManyArgumentsError(
                "t1, t2, and Theta delta specified. "
                "What do you want me to do??"
                )

        if passed == 0b1110:
            raise AnSecThetaTooManyArgumentsError(
                "t1, t2, and midTheta specified. "
                "What do you want me to do??"
                )

        if passed == 0b1111:
            raise AnSecThetaTooManyArgumentsError(
                "What do you want me to do??"
                )



import unittest

import numpy as np

import raimad as rai


class TestPolys(unittest.TestCase):

    def test_rectwh(self):
        rect = rai.RectWH(10, 20)
        self.assertIsNone(np.testing.assert_array_almost_equal(
            rect.geoms['root'][0],
            [
                [-5, -10],
                [5, -10],
                [5, 10],
                [-5, 10]
            ]))

    def test_circle(self):
        circle = rai.Circle(69)
        points = circle.geoms['root'][0]
        self.assertGreaterEqual(len(points), 10)
        for point in points:
            self.assertAlmostEqual(
                np.linalg.norm(point),
                69
                )

    def test_rectwire(self):
        rectwire = rai.RectWire((0, 0), (0, 20), 10)
        self.assertIsNone(np.testing.assert_array_almost_equal(
            rectwire.geoms['root'][0],
            [
                [-5, 0],
                [5, 0],
                [5, 20],
                [-5, 20]
            ]))

    def test_ansec(self):
        same = [
            rai.AnSec(
                r1=10,
                r2=20,
                theta1=rai.eigthcircle,
                theta2=-rai.eigthcircle,
                orientation=rai.Orientation.NEG
                ),
            rai.AnSec(
                r1=10,
                r2=20,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                orientation=rai.Orientation.POS
                ),
            rai.AnSec(
                r1=10,
                dr=10,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                orientation=rai.Orientation.POS
                ),
            rai.AnSec(
                r2=20,
                dr=10,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                orientation=rai.Orientation.POS
                ),
            rai.AnSec(
                r2=20,
                dr=10,
                theta1=-rai.eigthcircle,
                dtheta=rai.quartercircle,
                orientation=rai.Orientation.POS
                ),
            rai.AnSec(
                r2=20,
                dr=10,
                theta2=rai.eigthcircle,
                dtheta=rai.quartercircle,
                orientation=rai.Orientation.POS
                ),
            rai.AnSec(
                rmid=15,
                dr=10,
                theta2=-rai.eigthcircle,
                dtheta=-rai.quartercircle,
                orientation=rai.Orientation.NEG
                ),
            rai.AnSec(
                rmid=15,
                dr=10,
                thetamid=0,
                dtheta=rai.quartercircle,
                orientation=rai.Orientation.POS
                ),
            rai.AnSec(
                rmid=15,
                dr=-10,
                thetamid=0,
                dtheta=-rai.quartercircle,
                orientation=rai.Orientation.NEG
                ),
            ]

        geoms = [np.sort(ansec.geoms['root'][0], axis=0) for ansec in same]

        self.assertTrue(np.allclose(geoms[0], geoms))

        # invalid
        with self.assertRaises(rai.err.AnSecError):
            rai.AnSec(
                r2=20,
                r1=10,
                dr=-10,
                theta1=-rai.eigthcircle,
                theta2=rai.eigthcircle,
                orientation=rai.Orientation.POS
                )

        with self.assertRaises(rai.err.AnSecError):
            rai.AnSec(
                r2=20,
                dr=10,
                theta2=rai.eigthcircle,
                theta1=0,
                dtheta=-rai.quartercircle,
                orientation=rai.Orientation.NEG
                )

        with self.assertRaises(rai.err.AnSecError):
            rai.AnSec(
                rmid=15,
                theta2=rai.eigthcircle,
                dtheta=rai.quartercircle,
                orientation=rai.Orientation.NEG
                ),

        with self.assertRaises(rai.err.AnSecError):
            rai.AnSec(
                rmid=15,
                dr=5,
                dtheta=rai.semicircle,
                orientation=rai.Orientation.NEG
                ),


if __name__ == '__main__':
    unittest.main()




"""
Namespace flattening for RAIMAD.
"""

import numpy as np
Point = np.ndarray  # TODO

from raimad.empty import Empty
from raimad import graphviz as gv
from raimad.helpers import *
from raimad import affine
import raimad.iters as iters
from raimad.iters import (
    overlap,
    nonoverlap,

    duplets,
    triplets,
    quadlets,
    quintlets,

    couples,
    triples,
    quadles,
    quintles,

    flatten,
    braid
    )
from raimad.string_import import string_import
from raimad.docparse import split_docstring

from raimad.dictlist import DictList

from raimad.mark import Mark
from raimad.layer import Layer
from raimad.option import Option
from raimad.boundpoint import BoundPoint
from raimad.transform import Transform
from raimad.compo import Compo
from raimad.compo import MarksContainer
from raimad.compo import SubcompoContainer
from raimad.proxy import Proxy
from raimad.proxy import LMap
from raimad.partial import Partial
from raimad.bbox import BBox

from raimad.rectwh import RectWH
from raimad.rectwire import RectWire
from raimad.circle import Circle
from raimad.ansec import AnSec
from raimad.custompoly import CustomPoly

from raimad import typing

from raimad import cif
from raimad.svg import export_svg
from raimad import err
from raimad import debug

from raimad.checker.violations import (
    Viol,
    MarksViol,
    LenientViol,
    RAI412,
    RAI442
    )

from raimad.checker.checker import (
    check_compo,
    check_module,
    Flake8Checker
    )

from raimad.snowman import Snowman

# Mypy really wants __all__ to be present
__all__ = [
    'export_svg',
    'Transform',
    'Compo',
    'DictList',
    'Proxy',
    'BoundPoint',
    'MarksContainer',
    'SubcompoContainer',
    'BBox',
    'Partial',
    'LMap',
    ]





import raimad as rai

class CustomPolyException(Exception):
    pass

# TODO tests
class CustomPoly(rai.Compo):
    """
    Custom Polygon

    A polygon that you can defined on-the fly!
    #TODO add example
    """
    browser_tags = ["builtin", "polygon"]

    class Options:
        points = rai.Option(
            "A list of points for the polygon",
            browser_default=[
                [0, 0],
                [10, 0],
                [10, 20],
                [20, -20],
                ]
            )

    def _make(self, points):
        processed_points = []
        for point in points:
            match point:
                case (
                        str() as mark_name,
                        (float() | int(), float() | int()) as point
                        ):
                    self.marks[mark_name] = point

                case (float() | int(), float() | int()) as point:
                    # Very confusing,
                    # when it comes to type annotations,
                    # `int` is treated as a subset of `float`,
                    # but in structural pattern matching,
                    # an `int` is not a `float`.
                    pass

                case _:
                    raise CustomPolyException(
                        "I have no clue what you want me to do with "
                        f"`{point}`"
                        )
            processed_points.append(point)
            # TODO would be cool to make it possible to
            # specify coordinates relative to the previous
            # point as opposed to absolute.

        self.geoms.update({'root': [processed_points]})




from copy import deepcopy
from typing import Self

import numpy as np
import raimad as rai

class Transform:
    """
    Transformation: container for affine matrix
    """

    _affine: 'rai.typing.Affine'

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self._affine = np.identity(3)

    def transform_xyarray(
            self,
            poly: 'rai.typing.Poly | pc.typing.PolyArray'
            ) -> 'rai.typing.Poly | pc.typing.PolyArray':
        """
        Apply transformation to xyarray and return new transformed xyarray
        """
        return rai.affine.transform_xyarray(self._affine, poly)

    def transform_point(self, point: 'rai.typing.Point') -> 'pc.typing.Point':
        """
        Apply transformation to point and return new transformed point
        """
        return rai.affine.transform_point(self._affine, point)

    def compose(self, transform: Self) -> Self:
        """
        Apply a transform to this transform
        """
        if transform is not None:
            self._affine = transform._affine @ self._affine
        return self

    def get_translation(self) -> np.typing.NDArray[np.float64]:
        return rai.affine.get_translation(self._affine)

    def get_rotation(self) -> float:
        return rai.affine.get_rotation(self._affine)

    def get_shear(self) -> float:
        return rai.affine.get_shear(self._affine)

    def get_scale(self) -> tuple[np.float64, np.float64]:
        return rai.affine.get_scale(self._affine)

    def does_translate(self) -> bool:
        norm = np.linalg.norm(self.get_translation())
        return norm > 0.001  # TODO epsilon

    def does_rotate(self) -> 'rai.typing.Bool':
        return abs(self.get_rotation()) > 0.001  # TODO epsilon

    def does_shear(self) -> 'rai.typing.Bool':
        return abs(self.get_shear()) > 0.001  # TODO epsilon

    def does_scale(self) -> 'rai.typing.Bool':
        scale_x, scale_y = self.get_scale()
        # TODO epsilon
        return abs(1 - scale_x) > 0.001 or abs(1 - scale_y) > 0.001

    def __repr__(self):
        does_translate = self.does_translate()
        does_rotate = self.does_rotate()
        does_shear = self.does_shear()
        does_scale = self.does_scale()

        move_x, move_y = self.get_scale()
        rotation = self.get_rotation()
        shear = self.get_shear()
        scale_x, scale_y = self.get_scale()

        if True not in {does_translate, does_rotate, does_shear, does_scale}:
            # Could also test if affine == identity
            return "<Identity Transform>"

        return ''.join((
            "<Transform ",
            f"Move ({move_x:+.2f}, {move_y:+.2f}) "
                if does_translate else '',
            f"Rotate {np.rad2deg(rotation):.2f}) "
                if does_rotate else '',
            f"Shear {shear:.2f} "
                if does_shear else '',
            f"Scale ({scale_x:.2f}, {scale_y:.2f})"
                if does_scale else '',
            ">",
            ))

    def copy(self) -> Self:
        return deepcopy(self)

    # TODO typing.point
    # types defined in own files
    # then mergen in rai.typing
    # or just PointType, CompoClassType, etc
    def move(self, x=0, y: float = 0):
        if isinstance(x, rai.Point):
            x, y = x
        self._affine = rai.affine.move(x, y) @ self._affine
        return self

    def movex(self, x: float = 0) -> Self:
        self._affine = rai.affine.move(x, 0) @ self._affine
        return self

    def movey(self, y: float = 0) -> Self:
        self._affine = rai.affine.move(0, y) @ self._affine
        return self

    #def scale(
    #        self,
    #        x: float | rai.Point,  # TODO typing.point
    #        y: float | None = None,
    #        cx: float = 0,
    #        cy: float = 0,
    #        ) -> Self:

    #    if isinstance(x, rai.Point):
    #        x, y = x

    #    elif y is None:
    #        y = x

    #    self._affine = rai.affine.around(pc.affine.scale(x, y), cx, cy) @ self._affine
    #    return self
    def scale(self, x: float, y: float | None = None) -> Self:
        if y is None:
            y = x
        self._affine = rai.affine.scale(x, y) @ self._affine
        return self

    def rotate(
            self,
            angle: float,
            x: float | rai.Point = 0,
            y: float = 0
            ) -> Self:

        if isinstance(x, rai.Point):
            x, y = x

        self._affine = rai.affine.around(rai.affine.rotate(angle), x, y) @ self._affine

        return self

    def hflip(self, x: float = 0) -> Self:
        self._affine = rai.affine.around(rai.affine.scale(1, -1), 0, x) @ self._affine
        return self

    def vflip(self, y: float = 0) -> Self:
        self._affine = rai.affine.around(rai.affine.scale(-1, 1), y, 0) @ self._affine
        return self

    def flip(self, x: float = 0, y: float = 0) -> Self:
        self._affine = rai.affine.around(rai.affine.scale(-1, -1), x, y) @ self._affine
        return self

    def inverse(self):
        self._affine = np.linalg.inv(self._affine)
        return self

